<?php

namespace Drupal\d3_tagcloud\Plugin\Block;

use Drupal\Core\Block\BlockBase;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides the Leo Cloud block.
 *
 * @Block(
 *   id = "d3_tagcloud",
 *   admin_label = @Translation("D3 Tag Cloud"),
 *   category = @Translation("Tag Cloud for Taxonomy"),
 * )
 */
class TagCloudBlock extends BlockBase implements ContainerFactoryPluginInterface {

  /**
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The term storage.
   *
   * @var \Drupal\taxonomy\TermStorageInterface
   */
  protected $termStorage;

  /**
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Constructs a new TagCloudBlock object.
   *
   * @param array $configuration
   *    A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *    The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *    The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *    The entity type manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   *    The entity field manager.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *    The module handler.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entityTypeManager, EntityFieldManagerInterface $entityFieldManager, ModuleHandlerInterface $moduleHandler) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->entityTypeManager = $entityTypeManager;
    $this->termStorage = $entityTypeManager->getStorage('taxonomy_term');
    $this->entityFieldManager = $entityFieldManager;
    $this->moduleHandler = $moduleHandler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('module_handler')
    );
  }


  /**
   * {@inheritdoc}
   */
  public function build() {
    $config = $this->getConfiguration();
    $vid = $config['taxonomy'];
    $render = [];

    $query = $this->termStorage->getQuery()
      ->accessCheck()
      ->condition('vid', $vid);
    if (!empty($config['visibility_field'])) {
      $query->condition($config['visibility_field'], 1)
        ->execute();
    }
    $tids = $query->execute();
    $terms = $this->termStorage->loadMultiple($tids);

    /** @var \Drupal\taxonomy\TermInterface $term */
    foreach ($terms as $term) {
      $link = $term->toUrl()->toString();
      $importance = !empty($config['importance_field']) ? $term->get($config['importance_field'])->getString() : 1;
      $render[$term->id()] = [
        'tid' => $term->id(),
        'text' => $term->label(),
        'link' => $link,
        'importance' => $importance,
      ];
    }

    $this->moduleHandler->alter('d3_tagcloud_render', $render);

    return [
      '#markup' => ' ',
      '#attached' => [
        'library' => ['d3_tagcloud/leo-terms'],
        'drupalSettings' => [
          'leoTerms' => [
            'termsList' => array_values($render),
          ],
        ],
      ],
      '#attributes' => [
        'class' => ['d3-term-cloud'],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function blockForm($form, FormStateInterface $form_state) {
    $form = parent::blockForm($form, $form_state);
    $config = $this->getConfiguration();

    $vocabularies = $this->entityTypeManager->getStorage('taxonomy_vocabulary')->loadMultiple();
    $taxonomies = array_map(function ($vid) {
      return $vid->label();
    }, $vocabularies);

    $form['taxonomy'] = [
      '#type' => 'select',
      '#title' => new TranslatableMarkup('Select taxonomy'),
      '#description' => new TranslatableMarkup('Taxonomy to display'),
      '#options' => array_merge(['' => new TranslatableMarkup('-- Please select --')], $taxonomies),
      '#ajax' => [
        'callback' => [$this, 'callbackTaxonomy'],
        'event' => 'change',
        'wrapper' => 'settings-fields',
        'method' => 'replace',
        'progress' => [
          'type' => 'throbber',
          'message' => t('Verifying entry...'),
        ],
      ],
      '#attributes' => [],
      '#default_value' => $config['taxonomy'] ?? '',
      '#required' => TRUE,
    ];

    $taxonomy = $config['taxonomy'];
    $form['settings'] = $this->createSettingsFields($taxonomy);
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function createSettingsFields($vid) {
    $form = [
      '#type' => 'container',
      '#attributes' => ['id' => 'settings-fields'],
    ];
    $form['visibility_field'] = $this->createVisibilityField($vid);
    $form['importance_field'] = $this->createImportanceField($vid);
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function createImportanceField($vid) {
    $config = $this->getConfiguration();
    $options = $this->getOptions($vid);

    $form['importance_field'] = [
      '#type' => 'select',
      '#title' => new TranslatableMarkup('Importance field'),
      '#attributes' => [
        'name' => 'field_select_importance',
      ],
      '#prefix' => '<div id="edit-output">',
      '#suffix' => '</div>',
      '#value' => !empty($config['importance_field']) ? $config['importance_field'] : 0,
    ];

    if (!empty($vid)) {
      $form['importance_field']['#options'] = array_merge(['' => new TranslatableMarkup('-- Please select --')], $options);
    }
    else {
      $form['importance_field']['#options'] = [0 => $this->t('-- None --')];
    }

    return $form['importance_field'];
  }

  /**
   * {@inheritdoc}
   */
  public function createVisibilityField($vid) {
    $config = $this->getConfiguration();
    $options = $this->getOptions($vid);

    $form['visibility_field'] = [
      '#type' => 'select',
      '#title' => new TranslatableMarkup('Visibility field'),
      '#attributes' => [
        'name' => 'field_select_visibility',
      ],
      '#prefix' => '<div id="edit-output">',
      '#suffix' => '</div>',
      '#value' => !empty($config['visibility_field']) ? $config['visibility_field'] : 0,
    ];

    if (!empty($vid)) {
      $form['visibility_field']['#options'] = array_merge(['' => new TranslatableMarkup('-- Please select --')], $options);
    }
    else {
      $form['visibility_field']['#options'] = [0 => $this->t('-- None --')];
    }

    return $form['visibility_field'];
  }

  /**
   * {@inheritdoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state) {
    $field = $form_state->getUserInput()['field_select_visibility'];
    $this->setConfigurationValue('taxonomy', $form_state->getValue('taxonomy'));
    $this->setConfigurationValue('visibility_field', $field);

    $field = $form_state->getUserInput()['field_select_importance'];
    $this->setConfigurationValue('importance_field', $field);
  }

  /**
   * {@inheritdoc}
   */
  public function callbackTaxonomy(array &$form, FormStateInterface $form_state) {
    return $this->createSettingsFields($form_state->getValue('settings')['taxonomy']);
  }

  /**
   * Gets the field definitions for a specific bundle.
   *
   * @param string $bundle
   *   The bundle.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface[]
   *   The array of field definitions for the bundle, keyed by field name.
   */
  private function getFieldDefinitions(string $bundle) {
    return $this->entityFieldManager->getFieldDefinitions('taxonomy_term', $bundle);
  }

  /**
   * {@inheritdoc}
   */
  private function getOptions(string $vid) {
    $options = [];
    $fields = $this->getFieldDefinitions($vid);
    foreach ($fields as $field) {
      /** @var \Drupal\Core\Field\BaseFieldDefinition $field */
      $name = $field->getFieldStorageDefinition()->getName();
      if (!str_starts_with($name, 'field_') && $name != 'name') {
        continue;
      }
      $options[$name] = $field->getLabel();
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags() {
    return Cache::mergeTags(parent::getCacheTags(), ['taxonomy_term_list']);
  }

}
